J = require './utils'

# Utils

defaultFormat = /(\d{1,4})/g

cards = [
  {
      type: 'maestro'
      pattern: /^(5018|5020|5038|6304|6759|676[1-3])/
      format: defaultFormat
      length: [12..19]
      cvcLength: [3]
      luhn: true
  }
  {
      type: 'dinersclub'
      pattern: /^(36|38|30[0-5])/
      format: defaultFormat
      length: [14]
      cvcLength: [3]
      luhn: true
  }
  {
      type: 'laser'
      pattern: /^(6706|6771|6709)/
      format: defaultFormat
      length: [16..19]
      cvcLength: [3]
      luhn: true
  }
  {
      type: 'jcb'
      pattern: /^35/
      format: defaultFormat
      length: [16]
      cvcLength: [3]
      luhn: true
  }
  {
      type: 'unionpay'
      pattern: /^62/
      format: defaultFormat
      length: [16..19]
      cvcLength: [3]
      luhn: false
  }
  {
      type: 'discover'
      pattern: /^(6011|65|64[4-9]|622)/
      format: defaultFormat
      length: [16]
      cvcLength: [3]
      luhn: true
  }
  {
      type: 'mastercard'
      pattern: /^5[1-5]/
      format: defaultFormat
      length: [16]
      cvcLength: [3]
      luhn: true
  }
  {
      type: 'amex'
      pattern: /^3[47]/
      format: /(\d{1,4})(\d{1,6})?(\d{1,5})?/
      length: [15]
      cvcLength: [3..4]
      luhn: true
  }
  {
      type: 'visa'
      pattern: /^4/
      format: defaultFormat
      length: [13..16]
      cvcLength: [3]
      luhn: true
  }
]

cardFromNumber = (num) ->
  num = (num + '').replace(/\D/g, '')
  return card for card in cards when card.pattern.test(num)

cardFromType = (type) ->
  return card for card in cards when card.type is type

luhnCheck = (num) ->
  odd = true
  sum = 0

  digits = (num + '').split('').reverse()

  for digit in digits
    digit = parseInt(digit, 10)
    digit *= 2 if (odd = !odd)
    digit -= 9 if digit > 9
    sum += digit

  sum % 10 == 0

hasTextSelected = (target) ->
  # If some text is selected
  return true if target.selectionStart? and
    target.selectionStart isnt target.selectionEnd

  # If some text is selected in IE
  return true if document?.selection?.createRange?().text

  false

# Private

# Format Card Number

reFormatCardNumber = (e) ->
  setTimeout =>
    target = e.currentTarget
    value   = J.val(target)
    value   = Payment.fns.formatCardNumber(value)
    J.val(target, value)

formatCardNumber = (e) ->
  # Only format if input is a number
  digit = String.fromCharCode(e.which)
  return unless /^\d+$/.test(digit)

  target = e.currentTarget
  value   = J.val(target)
  card    = cardFromNumber(value + digit)
  length  = (value.replace(/\D/g, '') + digit).length

  upperLength = 16
  upperLength = card.length[card.length.length - 1] if card
  return if length >= upperLength

  # Return if focus isn't at the end of the text
  return if target.selectionStart? and
    target.selectionStart isnt value.length

  if card && card.type is 'amex'
    # Amex cards are formatted differently
    re = /^(\d{4}|\d{4}\s\d{6})$/
  else
    re = /(?:^|\s)(\d{4})$/

  # If '4242' + 4
  if re.test(value)
    e.preventDefault()
    J.val(target, value + ' ' + digit)

  # If '424' + 2
  else if re.test(value + digit)
    e.preventDefault()
    J.val(target, value + digit + ' ')

formatBackCardNumber = (e) ->
  target = e.currentTarget
  value   = J.val(target)

  return if e.meta

  # Return unless backspacing
  return unless e.which is 8

  # Return if focus isn't at the end of the text
  return if target.selectionStart? and
    target.selectionStart isnt value.length

  # Remove the trailing space
  if /\d\s$/.test(value)
    e.preventDefault()
    J.val(target, value.replace(/\d\s$/, ''))
  else if /\s\d?$/.test(value)
    e.preventDefault()
    J.val(target, value.replace(/\s\d?$/, ''))

# Format Expiry

formatExpiry = (e) ->
  # Only format if input is a number
  digit = String.fromCharCode(e.which)
  return unless /^\d+$/.test(digit)

  target = e.currentTarget
  val     = J.val(target) + digit

  if /^\d$/.test(val) and val not in ['0', '1']
    e.preventDefault()
    J.val(target, "0#{val} / ")

  else if /^\d\d$/.test(val)
    e.preventDefault()
    J.val(target, "#{val} / ")

formatForwardExpiry = (e) ->
  digit = String.fromCharCode(e.which)
  return unless /^\d+$/.test(digit)

  target = e.currentTarget
  val     = J.val(target)

  if /^\d\d$/.test(val)
    J.val(target, "#{val} / ")

formatForwardSlash = (e) ->
  slash = String.fromCharCode(e.which)
  return unless slash is '/'

  target = e.currentTarget
  val     = J.val(target)

  if /^\d$/.test(val) and val isnt '0'
    J.val(target, "0#{val} / ")

formatBackExpiry = (e) ->
  # If shift+backspace is pressed
  return if e.metaKey

  target = e.currentTarget
  value   = J.val(target)

  # Return unless backspacing
  return unless e.which is 8

  # Return if focus isn't at the end of the text
  return if target.selectionStart? and
    target.selectionStart isnt value.length

  # Remove the trailing space
  if /\d(\s|\/)+$/.test(value)
    e.preventDefault()
    J.val(target, value.replace(/\d(\s|\/)*$/, ''))
  else if /\s\/\s?\d?$/.test(value)
    e.preventDefault()
    J.val(target, value.replace(/\s\/\s?\d?$/, ''))

#  Restrictions

restrictNumeric = (e) ->
  # Key event is for a browser shortcut
  return true if e.metaKey or e.ctrlKey

  # If keycode is a space
  return e.preventDefault() if e.which is 32

  # If keycode is a special char (WebKit)
  return true if e.which is 0

  # If char is a special char (Firefox)
  return true if e.which < 33

  input = String.fromCharCode(e.which)

  # Char is a number or a space
  return e.preventDefault() if !/[\d\s]/.test(input)

restrictCardNumber = (e) ->
  target = e.currentTarget
  digit   = String.fromCharCode(e.which)
  return unless /^\d+$/.test(digit)

  return if hasTextSelected(target)

  # Restrict number of digits
  value = (J.val(target) + digit).replace(/\D/g, '')
  card  = cardFromNumber(value)

  if card
    e.preventDefault() unless value.length <= card.length[card.length.length - 1]
  else
    # All other cards are 16 digits long
    e.preventDefault() unless value.length <= 16

restrictExpiry = (e) ->
  target = e.currentTarget
  digit   = String.fromCharCode(e.which)
  return unless /^\d+$/.test(digit)

  return if hasTextSelected(target)

  value = J.val(target) + digit
  value = value.replace(/\D/g, '')

  return e.preventDefault() if value.length > 6

restrictCVC = (e) ->
  target = e.currentTarget
  digit   = String.fromCharCode(e.which)
  return unless /^\d+$/.test(digit)

  val     = J.val(target) + digit
  return e.preventDefault() unless val.length <= 4

setCardType = (e) ->
  target  = e.currentTarget
  val      = J.val(target)
  cardType = Payment.fns.cardType(val) or 'unknown'

  unless J.hasClass(target, cardType)
    allTypes = (card.type for card in cards)

    J.removeClass target, 'unknown'
    J.removeClass target, allTypes.join(' ')

    J.addClass target, cardType
    J.toggleClass target, 'identified', cardType isnt 'unknown'
    J.trigger target, 'payment.cardType', cardType

# Public

class Payment
  @fns:
    cardExpiryVal: (value) ->
      value = value.replace(/\s/g, '')
      [month, year] = value.split('/', 2)

      # Allow for year shortcut
      if year?.length is 2 and /^\d+$/.test(year)
        prefix = (new Date).getFullYear()
        prefix = prefix.toString()[0..1]
        year   = prefix + year

      month = parseInt(month, 10)
      year  = parseInt(year, 10)

      month: month, year: year
    validateCardNumber: (num) ->
      num = (num + '').replace(/\s+|-/g, '')
      return false unless /^\d+$/.test(num)

      card = cardFromNumber(num)
      return false unless card

      num.length in card.length and
        (card.luhn is false or luhnCheck(num))
    validateCardExpiry: (month, year) ->
      # Allow passing an object
      if typeof month is 'object' and 'month' of month
        {month, year} = month

      return false unless month and year

      month = J.trim(month)
      year  = J.trim(year)

      return false unless /^\d+$/.test(month)
      return false unless /^\d+$/.test(year)
      return false unless parseInt(month, 10) <= 12

      if year.length is 2
        prefix = (new Date).getFullYear()
        prefix = prefix.toString()[0..1]
        year   = prefix + year

      expiry      = new Date(year, month)
      currentTime = new Date

      # Months start from 0 in JavaScript
      expiry.setMonth(expiry.getMonth() - 1)

      # The cc expires at the end of the month,
      # so we need to make the expiry the first day
      # of the month after
      expiry.setMonth(expiry.getMonth() + 1, 1)

      expiry > currentTime
    validateCardCVC: (cvc, type) ->
      cvc = J.trim(cvc)
      return false unless /^\d+$/.test(cvc)

      if type
        # Check against a explicit card type
        cvc.length in cardFromType(type)?.cvcLength
      else
        # Check against all types
        cvc.length >= 3 and cvc.length <= 4
    cardType: (num) ->
      return null unless num
      cardFromNumber(num)?.type or null
    formatCardNumber: (num) ->
      card = cardFromNumber(num)
      return num unless card

      upperLength = card.length[card.length.length - 1]

      num = num.replace(/\D/g, '')
      num = num[0..upperLength]

      if card.format.global
        num.match(card.format)?.join(' ')
      else
        groups = card.format.exec(num)
        groups?.shift()
        groups?.join(' ')
  @restrictNumeric: (el) ->
    J.on el, 'keypress', restrictNumeric
  @cardExpiryVal: (el) ->
    Payment.fns.cardExpiryVal(J.val(el))
  @formatCardCVC: (el) ->
    Payment.restrictNumeric el
    J.on el, 'keypress', restrictCVC
    el
  @formatCardExpiry: (el) ->
    Payment.restrictNumeric el
    J.on el, 'keypress', restrictExpiry
    J.on el, 'keypress', formatExpiry
    J.on el, 'keypress', formatForwardSlash
    J.on el, 'keypress', formatForwardExpiry
    J.on el, 'keydown', formatBackExpiry
    el
  @formatCardNumber: (el) ->
    Payment.restrictNumeric el
    J.on el, 'keypress', restrictCardNumber
    J.on el, 'keypress', formatCardNumber
    J.on el, 'keydown', formatBackCardNumber
    J.on el, 'keyup', setCardType
    J.on el, 'paste', reFormatCardNumber
    el

Payment.J = J
module.exports = Payment
global.Payment = Payment